/*
 * fbgrab - takes screenshots using the framebuffer.
 *
 * (C) Gunnar Monell <gmo@linux.nu> 2002
 *
 * This program is free Software, see the COPYING file
 * and is based on Stephan Beyer's <fbshot@s-beyer.de> FBShot
 * (C) 2000.
 *
 * For features and differences, read the manual page.
 *
 * This program has been checked with "splint +posixlib" without
 * warnings. Splint is available from http://www.splint.org/ .
 * Patches and enhancements of fbgrab have to fulfill this too.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/dir.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <getopt.h>
#include <sys/vt.h>   /* to handle vt changing */
#include <png.h>      /* PNG lib */
#include <linux/fb.h> /* to handle framebuffer ioctls */
#include <sys/mman.h>
#include <linux/input.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <syslog.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>

/* Change this to whatever your daemon is called */
#define DAEMON_NAME "FBGrab.daemon"

/* Change this to the user under which to run */
#define RUN_AS_USER "root"

#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1


#define	VERSION	"1.0 beta 1"
#define MAX_LEN 512
#define UNDEFINED -1


static int read_framebuffer(char *device, size_t bytes, unsigned char *buf_p)
{
    int fd;
    if(-1 == (fd=open(device, O_RDONLY)))
    {
	syslog( LOG_NOTICE, "Error: Couldn't open %s.\n", device);
	return -1;

    }

    if (buf_p == NULL || read(fd, buf_p, bytes) != (ssize_t) bytes)
    {
	syslog( LOG_NOTICE,"Error: Not enough memory or data\n");
	close(fd);
	return -1;
    }
    close(fd);
    return 1;
}

static void convert1555to32(int width, int height,
			    unsigned char *inbuffer,
			    unsigned char *outbuffer)
{
    unsigned int i;

    for (i=0; i < (unsigned int) height*width*2; i+=2)
    {
	/* BLUE  = 0 */
	outbuffer[(i<<1)+0] = (inbuffer[i+1] & 0x7C) << 1;
	/* GREEN = 1 */
        outbuffer[(i<<1)+1] = (((inbuffer[i+1] & 0x3) << 3) |
			     ((inbuffer[i] & 0xE0) >> 5)) << 3;
	/* RED   = 2 */
	outbuffer[(i<<1)+2] = (inbuffer[i] & 0x1f) << 3;
	/* ALPHA = 3 */
	outbuffer[(i<<1)+3] = '\0';
    }
}

static void convert565to32(int width, int height,
			   unsigned char *inbuffer,
			   unsigned char *outbuffer)
{
    unsigned int i;

    for (i=0; i < (unsigned int) height*width*2; i+=2)
    {
	/* BLUE  = 0 */
	outbuffer[(i<<1)+0] = (inbuffer[i] & 0x1f) << 3;
	/* GREEN = 1 */
        outbuffer[(i<<1)+1] = (((inbuffer[i+1] & 0x7) << 3) |
			     (inbuffer[i] & 0xE0) >> 5) << 2;
        /* RED   = 2 */
	outbuffer[(i<<1)+2] = (inbuffer[i+1] & 0xF8);
	/* ALPHA = 3 */
	outbuffer[(i<<1)+3] = '\0';
    }
}

static void convert888to32(int width, int height,
			   unsigned char *inbuffer,
			   unsigned char *outbuffer)
{
    unsigned int i;

    for (i=0; i < (unsigned int) height*width; i++)
    {
	/* BLUE  = 0 */
	outbuffer[(i<<2)+0] = inbuffer[i*3+0];
	/* GREEN = 1 */
        outbuffer[(i<<2)+1] = inbuffer[i*3+1];
	/* RED   = 2 */
        outbuffer[(i<<2)+2] = inbuffer[i*3+2];
	/* ALPHA */
        outbuffer[(i<<2)+3] = '\0';
    }
}

static int write_PNG(unsigned char *outbuffer, char *filename,
				int width, int height, int interlace)
{
    int i;
    int bit_depth=0, color_type;
    png_bytep row_pointers[height];
    png_structp png_ptr;
    png_infop info_ptr;
    FILE *outfile = fopen(filename, "wb");

    for (i=0; i<height; i++)
	row_pointers[i] = outbuffer + i * 4 * width;

    if (!outfile)
    {
	syslog( LOG_NOTICE, "Error: Couldn't fopen %s.\n", filename);
	return -1;
    }

    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
	(png_voidp) NULL, (png_error_ptr) NULL, (png_error_ptr) NULL);

    if (!png_ptr)
    {
	syslog( LOG_NOTICE, "Error: Couldn't create PNG write struct.");
	return -1;
    }


    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
    {
	png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
	syslog( LOG_NOTICE, "Error: Couldn't create PNG info struct.");
	return -1;
    }

    png_init_io(png_ptr, outfile);

    png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);

    bit_depth = 8;
    color_type = PNG_COLOR_TYPE_RGB_ALPHA;
    png_set_invert_alpha(png_ptr);
    png_set_bgr(png_ptr);

    png_set_IHDR(png_ptr, info_ptr, width, height,
		 bit_depth, color_type, interlace,
		 PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    png_write_info(png_ptr, info_ptr);

    png_write_image(png_ptr, row_pointers);

    png_write_end(png_ptr, info_ptr);
    /* puh, done, now freeing memory... */
    png_destroy_write_struct(&png_ptr, &info_ptr);

    if (outfile != NULL)
	(void) fclose(outfile);
	sync();
    return 1;
}

static int convert_and_write(unsigned char *inbuffer, char *filename,
				int width, int height, int bits, int interlace)
{
    size_t bufsize = (size_t) width * height * 4;

    unsigned char *outbuffer = malloc(bufsize);

    if (outbuffer == NULL)
	syslog( LOG_NOTICE,"Not enough memory");

    memset(outbuffer, 0, bufsize);

    switch(bits)
    {
        case 15:
        convert1555to32(width, height, inbuffer, outbuffer);
        return write_PNG(outbuffer, filename, width, height, interlace);
        break;
        case 16:
        convert565to32(width, height, inbuffer, outbuffer);
        return write_PNG(outbuffer, filename, width, height, interlace);
        break;
        case 24:
        convert888to32(width, height, inbuffer, outbuffer);
        return write_PNG(outbuffer, filename, width, height, interlace);
        break;
        case 32:
        /* No conversion needed */
        return write_PNG(inbuffer, filename, width, height, interlace);
        break;
        default:
        syslog( LOG_NOTICE, "%d bits per pixel are not supported! ", bits);
        return -1;
    }

    (void) free(outbuffer);
    return 1;
}


/********
 * MAIN *
 ********/



static void child_handler(int signum)
{
    switch(signum) {
    case SIGALRM: exit(EXIT_FAILURE); break;
    case SIGUSR1: exit(EXIT_SUCCESS); break;
    case SIGCHLD: exit(EXIT_FAILURE); break;
    }
}

static void daemonize( const char *lockfile )
{
    pid_t pid, sid, parent;
    int lfp = -1;

    /* already a daemon */
    if ( getppid() == 1 ) return;

    /* Create the lock file as the current user */
    if ( lockfile && lockfile[0] ) {
        lfp = open(lockfile,O_RDWR|O_CREAT,0640);
        if ( lfp < 0 ) {
            syslog( LOG_ERR, "unable to create lock file %s, code=%d (%s)",
                    lockfile, errno, strerror(errno) );
            exit(EXIT_FAILURE);
        }
    }

    /* Drop user if there is one, and we were run as root */
    if ( getuid() == 0 || geteuid() == 0 ) {
        struct passwd *pw = getpwnam(RUN_AS_USER);
        if ( pw ) {
            syslog( LOG_NOTICE, "setting user to " RUN_AS_USER );
            setuid( pw->pw_uid );
        }
    }

    /* Trap signals that we expect to recieve */
    signal(SIGCHLD,child_handler);
    signal(SIGUSR1,child_handler);
    signal(SIGALRM,child_handler);

    /* Fork off the parent process */
    pid = fork();
    if (pid < 0) {
        syslog( LOG_ERR, "unable to fork daemon, code=%d (%s)",
                errno, strerror(errno) );
        exit(EXIT_FAILURE);
    }
    /* If we got a good PID, then we can exit the parent process. */
    if (pid > 0) {

        /* Wait for confirmation from the child via SIGTERM or SIGCHLD, or
           for two seconds to elapse (SIGALRM).  pause() should not return. */
        alarm(2);
        pause();

        exit(EXIT_FAILURE);
    }

    /* At this point we are executing as the child process */
    parent = getppid();

    /* Cancel certain signals */
    signal(SIGCHLD,SIG_DFL); /* A child process dies */
    signal(SIGTSTP,SIG_IGN); /* Various TTY signals */
    signal(SIGTTOU,SIG_IGN);
    signal(SIGTTIN,SIG_IGN);
    signal(SIGHUP, SIG_IGN); /* Ignore hangup signal */
    signal(SIGTERM,SIG_DFL); /* Die on SIGTERM */

    /* Change the file mode mask */
    umask(0);

    /* Create a new SID for the child process */
    sid = setsid();
    if (sid < 0) {
        syslog( LOG_ERR, "unable to create a new session, code %d (%s)",
                errno, strerror(errno) );
        exit(EXIT_FAILURE);
    }

    /* Change the current working directory.  This prevents the current
       directory from being locked; hence not being able to remove it. */
    if ((chdir("/")) < 0) {
        syslog( LOG_ERR, "unable to change directory to %s, code %d (%s)",
                "/", errno, strerror(errno) );
        exit(EXIT_FAILURE);
    }
    /* Redirect standard files to /dev/null */
    freopen( "/dev/null", "r", stdin);
    //freopen( "/dev/null", "w", stdout); // does not work

    freopen( "/dev/null", "w", stderr);

    /* Tell the parent process that we are A-okay */
    kill( parent, SIGUSR1 );
}

int FileExists(char * FileName)
{
	FILE *Fp;
	Fp = fopen(FileName,"rb");
	if (Fp)
	{
		fclose(Fp);
		return 1;
	}
	else
		return 0;
}

int GetFBBitDepth(char* device)
{
    int fd;
    struct fb_var_screeninfo screeninfo;
    if(-1 == (fd=open(device, O_RDONLY)))
    {
        syslog( LOG_NOTICE, "Error: Couldn't open %s.\n", device);
        return -1;

    }
    ioctl(fd, FBIOGET_VSCREENINFO, &screeninfo);
    close(fd);
    return screeninfo.bits_per_pixel;
}

int main(int argc, char **argv)
{
    unsigned char *buf_p;
    char *device = NULL;
    char outfile[MAX_LEN];
    char fttime[80];
    int  bitdepth=UNDEFINED, height=UNDEFINED, width=UNDEFINED;
    size_t buf_size;
    int interlace = PNG_INTERLACE_NONE;
    int done = 1;
    int usedatetime =0;
    int Teller = 1,Teller2 = 0;
    int fd;
    size_t rb;
    FILE *f;
    struct input_event ev[64];
    int yalv;
    int Keys[10],Mask[10];
    int screenshot = 0;
    time_t now;
    struct tm *timestruct;

    for(Teller = 0;Teller<10;Teller++)
    {
        Keys[Teller] = 0;
        Mask[Teller] = 0;
    }
    Teller = 1;

    mkdir("/mnt/sd/photo",S_IRWXO | S_IRWXU | S_IRWXG);
    sprintf(outfile,"/mnt/sd/photo/screenshots%d",BITDEPTH);
    mkdir(outfile, S_IRWXO | S_IRWXU | S_IRWXG);

    sprintf(outfile,"/mnt/sd/photo/screenshots/lastscreenshot%d",BITDEPTH);
    f = fopen(outfile,"r");
    if(f)
    {
        fscanf(f,"%d\n",&Teller);
        fclose(f);
    }

    f = fopen("./keyconfig","r");
    if(f)
    {
        fscanf(f,"I=%d\n",&Mask[0]);
        fscanf(f,"II=%d\n",&Mask[1]);
        fscanf(f,"Y=%d\n",&Mask[2]);
        fscanf(f,"A=%d\n",&Mask[3]);
        fscanf(f,"B=%d\n",&Mask[4]);
        fscanf(f,"X=%d\n",&Mask[5]);
        fscanf(f,"L=%d\n",&Mask[6]);
        fscanf(f,"R=%d\n",&Mask[7]);
        fscanf(f,"HOME=%d\n",&Mask[8]);
        fscanf(f,"JOYSTICK=%d\n",&Mask[9]);
        fscanf(f,"USEDATETIME=%d\n",&usedatetime);
        fclose(f);
    }

    openlog( DAEMON_NAME, LOG_PID, LOG_LOCAL5 );
    syslog( LOG_INFO, "starting" );

    daemonize( "/tmp/" DAEMON_NAME );


	memset(outfile, 0, MAX_LEN);

	device = "/dev/fb0";
    bitdepth = BITDEPTH;
	height = 240;
	width = 320;
	buf_size = width * height * (((unsigned int) bitdepth + 7) >> 3);
	buf_p = malloc(buf_size);
	if(buf_p )
	{
		memset(buf_p, 0, buf_size);
		if ((fd = open("/dev/input/event1", O_RDONLY)) < 0) {
			done=0;
		}

    		while(done == 1)
    		{
                rb = read(fd,ev,sizeof(struct input_event)*64);

                if (rb < (int) sizeof(struct input_event)) {
                    done=0;
                }

                for (yalv = 0;yalv < (int) (rb / sizeof(struct input_event));yalv++)
                {
                    if (EV_KEY == ev[yalv].type)
                            switch(ev[yalv].code)
                            {
                                case 0x128:Keys[0] = ev[yalv].value; break;
                                case 0x129:Keys[1] = ev[yalv].value; break;
                                case 0x123:Keys[2] = ev[yalv].value; break;
                                case 0x120:Keys[3] = ev[yalv].value; break;
                                case 0x122:Keys[4] = ev[yalv].value; break;
                                case 0x121:Keys[5] = ev[yalv].value; break;
                                case 0x124:Keys[6] = ev[yalv].value; break;
                                case 0x125:Keys[7] = ev[yalv].value; break;
                                case 0x126:Keys[8] = ev[yalv].value; break;
                                case 0x12A:Keys[9] = ev[yalv].value; break;
                            }
                }

                screenshot = 1;
                for(Teller2 = 0; Teller2<10;Teller2++)
                {
                    if(Mask[Teller2] > 0)
                        if(! (Keys[Teller2] > 0))
                        {
                            screenshot = 0;
                            break;
                        }
                    Keys[Teller2] = 0;
                }

                if(screenshot == 1)
                {
                    done = read_framebuffer(device, buf_size, buf_p);
                    if(done == 1)
                    {
                        if(usedatetime != 1)
                        {
                            sprintf(outfile,"/mnt/sd/photo/screenshots%d/screenshot%d.png",BITDEPTH,Teller);
                            while(FileExists(outfile) == 1)
                            {
                                Teller++;
                                sprintf(outfile,"/mnt/sd/photo/screenshots%d/screenshot%d.png",BITDEPTH,Teller);
                            }

                            sprintf(outfile,"/mnt/sd/photo/screenshots/lastscreenshot%d",BITDEPTH);
                            f = fopen(outfile,"w");
                            if(f)
                            {
                                fprintf(f,"%d\n",Teller);
                                fclose(f);
                            }
                        }
                        else
                        {
                            Teller2 = 1;
                            now = time(NULL);
                            timestruct = localtime(&now);

                            strftime(fttime, 80, "%Y-%m-%d %H-%M-%S",timestruct);
                            sprintf(outfile,"/mnt/sd/photo/screenshots%d/%s.png",BITDEPTH,fttime);
                            while(FileExists(outfile) == 1)
                            {
                                Teller2++;
                                sprintf(outfile,"/mnt/sd/photo/screenshots%d/%s (%d).png",BITDEPTH,fttime,Teller2);
                            }
                        }
                        done = convert_and_write(buf_p, outfile, width, height, bitdepth, interlace);
                    }
                }
                usleep(25);

		}
		(void) free(buf_p);
	}



    syslog( LOG_NOTICE, "terminated" );
    closelog();
    return 0;
}



